@@ -78,9 +78,8 @@ class Connection:
 
         (self._session, rc) = mimerapi.mimerBeginSession8(dsn, user, password)
         if rc:
-#            if rc == 90:
-#                (ec, ev) = (DatabaseError, (90, "Login Failure"))
-#            else:
+            if rc == 90:
+                rc = -25031     # Login failure
             (ec, ev) = get_mimerapi_exception(rc, self._session)
             mimerapi.mimerEndSession(self._session)
             self._session = None
@@ -225,6 +224,20 @@ class Connection:
         else:
             self.autocommitmode = False
 
+    def reset(self):
+        """
+        Reset the connection. Close all cursors and do rollback if a transaction
+        is running. Reset auto commit to default.
+        """
+        if self.__cursors:
+            for cur in self.__cursors:
+                cur.close()
+        
+        if self._transaction:
+            self.rollback()
+        
+        self.autocommit(False)
+
     def __raise_exception(self, rc):
         self.errorhandler(self, None, get_error_class(rc),
                           (rc, mimerpy_error[rc]))
