@@ -22,21 +22,85 @@
 
 from mimerpy.mimPyExceptionHandler import *
 import mimerapi
-import collections
+import collections, decimal, uuid
+from types import GeneratorType
+
+
+def _pythonSetDecimal(statement, cur_column, parameter):
+    if parameter is None:
+        rc = mimerapi.mimerSetString8(statement, cur_column, parameter)
+    else: 
+        rc = mimerapi.mimerSetString8(statement, cur_column, str(parameter))
+    return rc
+
+def _pythonSetDecimalF(statement, cur_column, parameter):
+    if mimerapi._level >= 2:
+        return _pythonSetDecimal(statement, cur_column, parameter)
+    return -25102
+
+def _pythonGetDecimal(statement, cur_column):
+    rc, val = mimerapi.mimerGetString8(statement, cur_column)
+    if rc >= 0 and val is not None:
+        val = decimal.Decimal(val)
+    return (rc, val)
+
+def _pythonGetDecimalF(statement, cur_column):
+    if mimerapi._level >= 2:
+        return _pythonGetDecimal(statement, cur_column)
+    return (-25102, None)
+
+def _pythonGetInt(statement, col):
+    if mimerapi._level >= 2:
+        rc, val = mimerapi.mimerGetString8(statement, col)
+        if rc >= 0 and val is not None:
+            return (0, int(val))
+        return (rc, None)
+    return mimerapi.mimerGetInt64(statement, col)
+
+def _pythonSetInt(statement, col, val):
+    if mimerapi._level >= 2 and val is not None:
+        return mimerapi.mimerSetString8(statement, col, str(val))
+    return mimerapi.mimerSetInt64(statement, col, val)
+
+def _pythonGetUUID(statement, col):
+    rc, val = mimerapi.mimerGetUUID(statement, col)
+    if rc >= 0 and val is not None:
+        return (0, uuid.UUID(bytes=val))
+    return (rc, None)
+
+def _pythonSetUUID(statement, col, val):
+    return mimerapi.mimerSetUUID(statement, col, val.bytes)
 
 def _define_funcs():
     global get_funcs
     global set_funcs
 
+    TYPEID_MASK = 0x40000000
+
     get_funcs = {1: mimerapi.mimerGetString8,
-                 2: mimerapi.mimerGetString8,
-                 3: mimerapi.mimerGetString8,
+                 2: _pythonGetDecimal,
+                 31: _pythonGetDecimal,
+                 3: _pythonGetInt,
+                 4: _pythonGetDecimalF,
                  6: mimerapi.mimerGetInt32,
                  10: mimerapi.mimerGetDouble,
                  11: mimerapi.mimerGetString8,
                  12: mimerapi.mimerGetString8,
                  13: mimerapi.mimerGetString8,
                  14: mimerapi.mimerGetString8,
+                 15: mimerapi.mimerGetString8,
+                 16: mimerapi.mimerGetString8,
+                 17: mimerapi.mimerGetString8,
+                 18: mimerapi.mimerGetString8,
+                 19: mimerapi.mimerGetString8,
+                 20: mimerapi.mimerGetString8,
+                 21: mimerapi.mimerGetString8,
+                 22: mimerapi.mimerGetString8,
+                 23: mimerapi.mimerGetString8,
+                 24: mimerapi.mimerGetString8,
+                 25: mimerapi.mimerGetString8,
+                 26: mimerapi.mimerGetString8,
+                 27: mimerapi.mimerGetString8,
                  34: mimerapi.mimerGetBinary,
                  35: mimerapi.mimerGetBinary,
                  39: mimerapi.mimerGetString8,
@@ -47,20 +111,37 @@ def _define_funcs():
                  52: mimerapi.mimerGetInt64,
                  63: mimerapi.mimerGetString8,
                  56: mimerapi.mimerGetDouble,
-                 54: mimerapi.mimerGetDouble,
+                 54: mimerapi.mimerGetFloat,
                  57: mimerapi.mimerGetBlobData,
                  58: mimerapi.mimerGetNclobData8,
-                 59: mimerapi.mimerGetNclobData8}
+                 59: mimerapi.mimerGetNclobData8,
+                 8104: _pythonGetUUID,
+    }
 
     set_funcs = {1: mimerapi.mimerSetString8,
-                 2: mimerapi.mimerSetString8,
-                 3: mimerapi.mimerSetString8,
+                 2: _pythonSetDecimal,
+                 31: _pythonSetDecimal,
+                 3: _pythonSetInt,
+                 4: _pythonSetDecimalF,
                  6: mimerapi.mimerSetInt32,
                  10: mimerapi.mimerSetDouble,
                  11: mimerapi.mimerSetString8,
                  12: mimerapi.mimerSetString8,
                  13: mimerapi.mimerSetString8,
                  14: mimerapi.mimerSetString8,
+                 15: mimerapi.mimerSetString8,
+                 16: mimerapi.mimerSetString8,
+                 17: mimerapi.mimerSetString8,
+                 18: mimerapi.mimerSetString8,
+                 19: mimerapi.mimerSetString8,
+                 20: mimerapi.mimerSetString8,
+                 21: mimerapi.mimerSetString8,
+                 22: mimerapi.mimerSetString8,
+                 23: mimerapi.mimerSetString8,
+                 24: mimerapi.mimerSetString8,
+                 25: mimerapi.mimerSetString8,
+                 26: mimerapi.mimerSetString8,
+                 27: mimerapi.mimerSetString8,
                  34: mimerapi.mimerSetBinary,
                  35: mimerapi.mimerSetBinary,
                  39: mimerapi.mimerSetString8,
@@ -71,11 +152,13 @@ def _define_funcs():
                  52: mimerapi.mimerSetInt64,
                  63: mimerapi.mimerSetString8,
                  56: mimerapi.mimerSetDouble,
-                 54: mimerapi.mimerSetDouble,
+                 54: mimerapi.mimerSetFloat,
                  57: mimerapi.mimerSetBlobData,
                  58: mimerapi.mimerSetNclobData8,
                  59: mimerapi.mimerSetNclobData8,
-                 501: mimerapi.mimerSetNull}
+                 501: mimerapi.mimerSetNull,
+                 8104: _pythonSetUUID,
+    }
 
 _define_funcs()
 
@@ -89,8 +172,7 @@ class Cursor:
 
     def __init__(self, connection, session):
         """
-            Creates a cursor . For more information please visit
-            http://developer.mimer.com/python.
+            Creates a MimerPy cursor.
 
             session
                 reference to MimerSession sessionhandle.
@@ -107,6 +189,7 @@ class Cursor:
         self._number_of_columns = None
         self._last_query = None
         self._DDL_rc_value = None
+        self._column_type = []
 
         self.__session = session
         self.__statement = None
@@ -164,14 +247,12 @@ class Cursor:
         if (len(arg) > 1):
             if (isinstance(arg[1], dict)):
                 parameter_markers = arg[1]
-            # this is the problem, should be more strict on what the input can be - Erik 2018-10
             elif (not isinstance(arg[1], tuple) and (not isinstance(arg[1], list))):
                 parameter_markers = [arg[1]]
             else:
                 parameter_markers = arg[1]
         query = arg[0]
 
-
         # If same query is used twice there is not need for a new statement
         if (query != self._last_query or self.__mimcursor):
             self.__close_statement()
@@ -203,38 +284,49 @@ class Cursor:
                 self._number_of_parameters = rc_value
                 try:
 
-                    if (len(parameter_markers) != self._number_of_parameters):
+                    if (len(parameter_markers) < self._number_of_parameters):
                         self.__raise_exception(-25013)
 
                     # Column number starts a 1
                     for cur_column in range(1, self._number_of_parameters + 1):
-                        parameter_type = mimerapi.mimerParameterType(self.__statement, cur_column)
-                        self.__check_mimerapi_error(parameter_type, self.__statement)
+                        parameter_type = mimerapi.mimerParameterType(
+                            self.__statement, cur_column)
+                        self.__check_mimerapi_error(
+                            parameter_type, self.__statement)
 
                         if (isinstance(parameter_markers, dict)):
-                            rc_value, parameter_name = mimerapi.mimerParameterName8(self.__statement, cur_column)
-                            self.__check_mimerapi_error(rc_value, self.__statement)
+                            rc_value, parameter_name = mimerapi.mimerParameterName8(
+                                self.__statement, cur_column)
+                            self.__check_mimerapi_error(
+                                rc_value, self.__statement)
                             if parameter_name in parameter_markers:
-                                parameter = parameter_markers.get(parameter_name)
+                                parameter = parameter_markers.get(
+                                    parameter_name)
                                 if (parameter == None):
                                     parameter_type = 501
-                                rc_value = set_funcs[parameter_type](self.__statement, cur_column, parameter)
+                                rc_value = set_funcs[parameter_type](
+                                    self.__statement, cur_column, parameter)
                             else:
-                                self.__raise_exception(-25012,
-                                                       str(parameter_name))
+                                # self.__raise_exception(-25012,str(parameter_name))
+                                pass  # Skipping keys in dictionary that does not match with any column
                         else:
                             # If the parameter marker is None, we use mimerSetNull
-                            if (parameter_markers[cur_column - 1] == None):
-                                parameter_type = 501
+                            try:
+                                if (parameter_markers[cur_column - 1] == None):
+                                    parameter_type = 501
+                            except TypeError:
+                                # End up here when invalid parameters are used
+                                self.__raise_exception(-25013)
                             rc_value = set_funcs[parameter_type](self.__statement,
-                                                                       cur_column, parameter_markers[cur_column - 1])
+                                                                cur_column, parameter_markers[cur_column - 1])
                         self.__check_mimerapi_error(rc_value, self.__statement)
 
                 # Catching error for errorhandler
                 except KeyError as e:
-                    self.__raise_exception(-25020, exception=e) # &&&& ??
+                    self.__raise_exception(-25020, exception=e)  # &&&& ??
                 # Catching error for errorhandler
                 except TypeError as e:
+                    # End up here when invalid parameters are used¨
                     self.__raise_exception(-25020, exception=e)
                 # Catching error for errorhandler
                 except OverflowError as e:
@@ -262,14 +354,17 @@ class Cursor:
                 description = collections.namedtuple('Column_description',
                                                      'name type_code display_size internal_size precision scale null_ok')
                 self.description = ()
-
+                self._column_type = []
                 for cur_column in range(1, self._number_of_columns + 1):
-                    func_tuple = mimerapi.mimerColumnName8(self.__statement, cur_column)
+                    func_tuple = mimerapi.mimerColumnName8(
+                        self.__statement, cur_column)
                     rc_value = func_tuple[0]
                     self.__check_mimerapi_error(rc_value, self.__statement)
                     name = func_tuple[1]
-                    rc_value = mimerapi.mimerColumnType(self.__statement, cur_column)
+                    rc_value = mimerapi.mimerColumnType(
+                        self.__statement, cur_column)
                     self.__check_mimerapi_error(rc_value, self.__statement)
+                    self._column_type.append(rc_value)
                     type_code = rc_value
                     self.description = self.description + (description(name=name,
                                                                        type_code=type_code,
@@ -295,17 +390,24 @@ class Cursor:
         """
         self.__check_if_open()
         self.__check_for_transaction()
-        rc_value = 0
         self._last_query = None
+        self.rowcount = 0
+        rc_value = 0
         values = []
 
+        if isinstance(params, GeneratorType):
+            tmp_params = []
+            for i in params:
+                tmp_params.append(i)
+            params = tmp_params
+
         # I would like to look over this at some point
         # Checking for invalid parameter structure
         if (not isinstance(params, tuple) and not isinstance(params, list)):
             self.__raise_exception(-25013)
-        else:
-            if (not isinstance(params[0], tuple) and not isinstance(params[0], dict)):
-                self.__raise_exception(-25013)
+        # else:
+        #    if (not isinstance(params[0], tuple) and not isinstance(params[0], dict)):
+        #        self.__raise_exception(-25013)
 
         self.__close_statement()
         values = mimerapi.mimerBeginStatement8(self.__session, query, 0)
@@ -317,8 +419,6 @@ class Cursor:
 
         rc_value = mimerapi.mimerParameterCount(self.__statement)
         self._number_of_parameters = rc_value
-
-        self.rowcount = 0
         self.__check_mimerapi_error(rc_value, self.__statement)
 
         try:
@@ -327,25 +427,34 @@ class Cursor:
 
                 # Column number starts a 1
                 for cur_column in range(1, self._number_of_parameters + 1):
-                    parameter_type = mimerapi.mimerParameterType(self.__statement, cur_column)
-                    self.__check_mimerapi_error(parameter_type, self.__statement)
+                    parameter_type = mimerapi.mimerParameterType(
+                        self.__statement, cur_column)
+                    self.__check_mimerapi_error(
+                        parameter_type, self.__statement)
 
                     if (isinstance(cur_param, dict)):
-                        rc_value, parameter_name = mimerapi.mimerParameterName8(self.__statement, cur_column)
+                        rc_value, parameter_name = mimerapi.mimerParameterName8(
+                            self.__statement, cur_column)
                         self.__check_mimerapi_error(rc_value, self.__statement)
                         if parameter_name in cur_param:
                             parameter = cur_param.get(parameter_name)
                             if (parameter == None):
                                 parameter_type = 501
-                            rc_value = set_funcs[parameter_type](self.__statement, cur_column, parameter)
+                            rc_value = set_funcs[parameter_type](
+                                self.__statement, cur_column, parameter)
                         else:
-                            self.__raise_exception(-25012, str(parameter_name))
+                            #self.__raise_exception(-25012, str(parameter_name))
+                            pass  # Skipping keys in dictionary that does not match with any column
                     else:
                         # If the parameter marker is None, we use mimerSetNull
-                        if (cur_param[cur_column - 1] == None):
-                            parameter_type = 501
+                        try:
+                            if (cur_param[cur_column - 1] == None):
+                                parameter_type = 501
+                        except TypeError as e:
+                            # End up here when invalid parameters are used
+                            self.__raise_exception(-25013)
                         rc_value = set_funcs[parameter_type](self.__statement,
-                                                                   cur_column, cur_param[cur_column - 1])
+                                                             cur_column, cur_param[cur_column - 1])
                 self.messages = []
 
                 # Batching after all parameters are set
@@ -389,9 +498,8 @@ class Cursor:
             return []
 
         for cur_column in range(1, self._number_of_columns + 1):
-            rc_value = mimerapi.mimerColumnType(self.__statement, cur_column)
-            self.__check_mimerapi_error(rc_value, self.__statement)
-            func_tuple = get_funcs[rc_value](self.__statement, cur_column)
+            func_tuple = get_funcs[self._column_type[cur_column - 1]
+                                   ](self.__statement, cur_column)
             self.__check_mimerapi_error(func_tuple[0], self.__statement)
 
             # Conversion from C int to Python boolean
@@ -402,7 +510,6 @@ class Cursor:
                     return_tuple = return_tuple + (True,)
             else:
                 return_tuple = return_tuple + (func_tuple[1],)
-
         return return_tuple
 
     def fetchmany(self, *arg):
@@ -435,14 +542,15 @@ class Cursor:
         fetch_length = self.arraysize
         rc_value = mimerapi.mimerFetch(self.__statement)
         fetch_value = rc_value
+
         while (fetch_value != 100 and fetch_length > 0):
             self.__check_mimerapi_error(fetch_value, self.__statement)
             return_tuple = ()
+
             # Column number starts a 1
             for cur_column in range(1, self._number_of_columns + 1):
-                rc_value = mimerapi.mimerColumnType(self.__statement, cur_column)
-                self.__check_mimerapi_error(rc_value, self.__statement)
-                func_tuple = get_funcs[rc_value](self.__statement, cur_column)
+                func_tuple = get_funcs[self._column_type[cur_column - 1]
+                                       ](self.__statement, cur_column)
                 self.__check_mimerapi_error(func_tuple[0], self.__statement)
 
                 # Conversion from C int to Python boolean
@@ -477,14 +585,15 @@ class Cursor:
         values = []
         rc_value = mimerapi.mimerFetch(self.__statement)
         fetch_value = rc_value
+
         while (fetch_value != 100):
             self.__check_mimerapi_error(fetch_value, self.__statement)
             return_tuple = ()
+
             # Column number starts a 1
             for cur_column in range(1, self._number_of_columns + 1):
-                rc_value = mimerapi.mimerColumnType(self.__statement, cur_column)
-                self.__check_mimerapi_error(rc_value, self.__statement)
-                func_tuple = get_funcs[rc_value](self.__statement, cur_column)
+                func_tuple = get_funcs[self._column_type[cur_column - 1]
+                                       ](self.__statement, cur_column)
                 self.__check_mimerapi_error(func_tuple[0], self.__statement)
 
                 # Conversion from C int to Python boolean
@@ -495,7 +604,6 @@ class Cursor:
                         return_tuple = return_tuple + (True,)
                 else:
                     return_tuple = return_tuple + (func_tuple[1],)
-
             values.append(return_tuple)
             fetch_value = mimerapi.mimerFetch(self.__statement)
         return values
@@ -522,7 +630,7 @@ class Cursor:
     def __close_statement(self):
         # Private method for closing MimerStatement.
         if (self.__statement is not None and
-            self.connection._session is not None):
+                self.connection._session is not None):
             rc_value = mimerapi.mimerEndStatement(self.__statement)
             self.__check_mimerapi_error(rc_value, self.__statement)
         self.__statement = None
@@ -560,6 +668,7 @@ class Cursor:
         self.__raise_exception(-25000)
 
 
+
 class ScrollCursor(Cursor):
     """
         Subclass to the Cursor-class where the cursor can be scrolled to
